
class Ram(object):
    def __init__(self, instr, input_registers):
        self.instr = instr # list of instructions
        self.current = 0 # current instruction
        self.input_registers = input_registers # I registers
        self.output_registers = [] # O registers
        self.work_registers = [] # R registers

    def read_register(self, type_register, index, ref_origin=None, ref_target=None):
        '''
        Read from a register
        '''
        self.current += 1
        if type_register == "input":
            return self.input_registers[index]
        elif type_register == "work":
            value = self.work_registers[index]
            if value is None:
                raise ValueError("register empty")
            else:
                return self.work_registers[index]
        elif type_register == "output":
            raise TypeError("output registers are write only")
        elif type_register == "value":
            return index
        elif type_register == "reference":
            target_index = self.read_register(ref_origin, index)
            return self.read_register(ref_target, target_index)

    def write_register(self, value, type_register, index, ref_origin=None, ref_target=None):
        '''
        Write to a register
        '''
        self.current += 1
        if type_register == "input":
            raise TypeError("output registers are read only")
        elif type_register == "work":
            if len(self.work_registers) <= index:
                self.work_registers.extend([None] * (index + 1 - len(self.work_registers))) # extend with uninitialized values
            self.work_registers[index] = value
        elif type_register == "output":
            if len(self.output_registers) <= index:
                self.output_registers.extend([None] * (index + 1 - len(self.output_registers))) # extend with uninitialized values
            self.output_registers[index] = value
        elif type_register == "value":
            raise TypeError("cannot write on value")
        elif type_register == "reference":
            target_index = self.read_register(ref_origin, index)
            self.write_register(ref_target, target_index, value)

    def op(self, type_op, r1, r2, r3):
        if type_op == 'add':
            value_1 = self.read_register(r1)
            value_2 = self.read_register(r2)
            value_3 = value_1 + value_2
            self.write_register(value_3, r3)
        elif type_op == 'sub':
            value_1 = self.read_register(r1)
            value_2 = self.read_register(r2)
            value_3 = value_1 - value_2
            self.write_register(value_3, r3)
        elif type_op == 'div':
            value_1 = self.read_register(r1)
            value_2 = self.read_register(r2)
            value_3 = value_1 / value_2
            self.write_register(value_3, r3)
        elif type_op == 'mul':
            value_1 = self.read_register(r1)
            value_2 = self.read_register(r2)
            value_3 = value_1 * value_2
            self.write_register(value_3, r3)

    def op_ctrl(self, type_op, z, r1=None, r2=None):
        if type_op == 'jump':
            if isinstance(z, int):
                self.current += z
            else:
               raise ValueError("wrong operand type (should be integer)")
        if type_op == 'je':
            value_1 = self.read_register(r1)
            value_2 = self.read_register(r2)
            if value_1 == value_2:
                self.op_ctrl('jump', z)
            else:
                self.current += 1
        if type_op == 'jl':
            value_1 = self.read_register(r1)
            value_2 = self.read_register(r2)
            if value_1 > value_2:
                self.op_ctrl('jump', z)
            else:
                self.current += 1

